# -*- coding: utf-8 -*-

#  Gedit message receiver plugin
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os

from gi.repository import Gedit
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import Pango

import dbus
import dbus.service
import dbus.mainloop.glib


if sys.version_info[0] < 3:
    str2or3 = lambda s: s.encode('utf-8')
else:
    str2or3 = lambda s: s
    
def escape(text):
    text = str2or3(text)
    return GObject.markup_escape_text(text)


class TracebackPanel:
    """
    Gets a python traceback (and related info) and handles:
    - creating a result panel if not exists
    - displaying traceback
    A TracebackPanel object lives until its result panel is closed.
    """
    def __init__ (self, panel):
        self.show_existing_only = False
        self.show_home_only = False
        
        self.first_line_time = 0
        
        ui_file = os.path.join(os.path.dirname(__file__), "python-traceback.ui")
        self.builder = Gtk.Builder()
        self.builder.add_from_file(ui_file)
        self.result_container = self.builder.get_object('hbox_tracebacks')
        self.scrolledwindow = self.builder.get_object('scrolledwindow_tracebacks')
        
        icon = Gtk.Image.new_from_icon_name("python", Gtk.IconSize.MENU)
        panel.add_item(self.result_container, 'python-traceback', 'Python Traceback', icon)
        panel.activate_item(self.result_container)
        
        panel.set_property("visible", True)
        
        # timestamp, display_text, filename, lineno, offset
        self.treestore = Gtk.TreeStore(float, str, str, int, int)
        treesort = self.treestore.sort_new_with_model()
        treesort.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.treeview = self.builder.get_object('treeview_result')
        self.treeview.set_model(treesort)
        
        self.treeview.set_search_equal_func(result_search_cb, None)
        
        tc = Gtk.TreeViewColumn("File", Gtk.CellRendererText(), markup=1)
        self.treeview.append_column(tc)
        
        button = self.builder.get_object('checkbutton_existing_only')
        button.set_active(self.show_existing_only)
        button = self.builder.get_object('checkbutton_home_only')
        button.set_active(self.show_home_only)
        
    def is_show_file(self, filename):
        if filename.startswith('/') and os.path.exists(filename):
            return not self.show_home_only or filename.startswith(os.getenv('HOME')+'/')
        else:
            return not self.show_existing_only
        
    def append_traceback(self, timestamp, exc_type, exc_value, exc_traceback, filename, lineno, offset):
        if self.first_line_time > timestamp:
            return
        adj = self.scrolledwindow.get_vadjustment()
        rows = []
        for i, (filename_tb, lineno_tb, func_name, text) in enumerate(exc_traceback):
            if self.is_show_file(filename_tb):
                rows.append([float(i), "<b>%d</b> in <b>%s</b>: %s"
                                    % (lineno_tb, escape(func_name), escape(text)),
                            str2or3(filename_tb), int(lineno_tb), 0])
        if rows or self.is_show_file(filename):
            display_text = '%s: <b>%s</b>' % (escape(exc_type), escape(exc_value))
            it = self.treestore.append(None,
                        [float(timestamp), display_text, str(filename), int(lineno), int(offset)])
            path = self.treestore.get_path(it)
            for row in rows:
                self.treestore.append(it, row)
            if adj.props.value + adj.props.page_size >= adj.props.upper:
                self.treeview.expand_row(path, False)
                self.treeview.set_cursor(path, None, False)
                GObject.idle_add(self.scroll_to_tb, path)
        
    def clear_tracebacks(self, timestamp):
        self.first_line_time = timestamp
        for row in reversed(self.treestore):
            if row[0] < timestamp:
                self.treestore.remove(row.iter)
                
    def scroll_to_tb(self, path):
        self.treeview.scroll_to_cell(path, None, True, 0.0)
        
    def destroy(self):
        self.result_container = None
        self.treestore = None
        self.treeview = None
        self.builder = None
        
    def on_button_clear_clicked(self, unused_button):
        self.treestore.clear()
    def on_checkbutton_existing_only_toggled (self, button):
        self.show_existing_only = button.get_active()
    def on_checkbutton_home_only_toggled (self, button):
        self.show_home_only = button.get_active()

    def on_treeview_result_button_press_event(self, treeview, event):
        if event.button == 3:
            path_column_cellxy = treeview.get_path_at_pos(int(event.x), int(event.y))
            if path_column_cellxy != None:
                treeview.grab_focus()
                treeview.set_cursor(path_column_cellxy[0], path_column_cellxy[1], False)

                menu = Gtk.Menu()
                mi = Gtk.ImageMenuItem.new_from_stock("gtk-copy", None)
                mi.connect_object("activate", TracebackPanel.on_menuitem_copy_activate,
                                                self, treeview, path_column_cellxy[0])
                mi.show()
                menu.append(mi)
                
                menu.attach_to_widget(treeview, None)
                menu.popup(None, None, None, None, event.button, event.time)
                return True
        else:
            return False

    def on_menuitem_copy_activate(self, treeview, path):
        it = treeview.get_model().get_iter(path)
        markup_text = treeview.get_model().get_value(it, 1)
        plain_text = Pango.parse_markup(markup_text, -1, u'\x00')[2]

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(plain_text, len(plain_text))
        clipboard.store()


def result_search_cb(model, unused_column, key, it, unused_data):
    """Callback function for searching in result list"""
    line_text = model.get_value(it, 1) # search in col 1 message
    plain_text = Pango.parse_markup(line_text, -1, u'\x00')[2] # remove Pango markup
    
    # if search text contains only lower-case characters, do case-insensitive matching:
    if key.islower():
        plain_text = plain_text.lower()
    
    # if the line contains the search text, it matches:
    if plain_text.find(key) >= 0:
        return False
    
    # line doesn't match:
    return True


class TracebackReceiver(dbus.service.Object):
    def __init__(self, bus, name, app):
        dbus.service.Object.__init__(self, bus, name)
        self._app = app
    
    @dbus.service.method("net.launchpad.PyTracebackReceiverInterface", in_signature='dssa(siss)sii')
    def Display(self, timestamp, exc_type, exc_value, exc_traceback, filename, lineno, offset):
        if self._app.current_window:
            GObject.idle_add(self._app.current_window.traceback_received,
                                timestamp,
                                exc_type, exc_value, exc_traceback,
                                filename, lineno, offset)
    
    @dbus.service.method("net.launchpad.PyTracebackReceiverInterface", in_signature='d')
    def Clear(self, timestamp):
        if self._app.current_window:
            GObject.idle_add(self._app.current_window.clear_received, timestamp)
    

class TracebackAppActivatable(GObject.Object, Gedit.AppActivatable):
    __gtype_name__ = "TracebackAppActivatable"
    app = GObject.property(type=Gedit.App)
    
    def __init__(self):
        GObject.Object.__init__(self)
        
        self.current_window = None
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self._bus_name = "net.launchpad.gedit-python-traceback"
        dbus.validate_bus_name(self._bus_name)
        
    def do_activate(self):
        self.current_window = None
        self._session_bus = dbus.SessionBus()
        self._service_bus_name = dbus.service.BusName(self._bus_name, self._session_bus)
        self._dbus_object = TracebackReceiver(self._session_bus, '/TBR', self)
        TracebackWindowActivatable.app = self
        
    def do_deactivate(self):
        self.current_window = None
        TracebackWindowActivatable.app = None
        self._dbus_object.remove_from_connection()
        self._dbus_object = None
        self._service_bus_name = None
        self._session_bus = None
        
        
class TracebackWindowActivatable(GObject.Object, Gedit.WindowActivatable):
    __gtype_name__ = "TracebackWindowActivatable"
    window = GObject.property(type=Gedit.Window)
    app = None
    
    def __init__(self):
        GObject.Object.__init__(self)
        self._panel = None
        
    def do_activate(self):
        if self.app:
            self.app.current_window = self
            
    def do_deactivate(self):
        if self.app:
            self.app.current_window = None
        if self._panel:
            self.destroy_panel()
            
    def destroy_panel(self):
        panel = self.window.get_bottom_panel()
        panel.remove_item(self._panel.result_container)
        self._panel.destroy()
        self._panel = None
        
    def do_update_state(self):
        if self.app:
            self.app.current_window = self
            
    def traceback_received(self, timestamp, exc_type, exc_value, exc_traceback, filename, lineno, offset):
        if self._panel is None:
            def connect_func(builder, widget, signal_name, handler_name, connect_object, flags, user_data):
                assert connect_object is None
                assert user_data is None
                assert flags == 0
                if hasattr(self, handler_name):
                    instance = self
                elif hasattr(self._panel, handler_name):
                    instance = self._panel
                else:
                    assert False
                widget.connect(signal_name, getattr(instance, handler_name))
            panel = self.window.get_bottom_panel()
            self._panel = TracebackPanel(panel)
            self._panel.builder.connect_signals_full(connect_func, None)
        self._panel.append_traceback(timestamp, exc_type, exc_value, exc_traceback,
                                        filename, lineno, offset)
        
    def clear_received(self, timestamp):
        if self._panel is not None:
            self._panel.clear_tracebacks(timestamp)
            
    def on_button_close_clicked(self, unused_button):
        self.destroy_panel()
        
    def on_treeview_result_row_activated(self, treeview, path, unused_col):
        model = treeview.get_model()
        filename = model[path][2]
        line_pos = model[path][3]
        col_pos = model[path][4]
        
        if filename and os.path.exists(filename):
            location = Gio.file_new_for_path(filename)
            tab = self.window.get_tab_from_location(location)
            if tab is None:
                #XXX: create_tab_from_location(..., False, True) crashes gedit
                tab = self.window.create_tab_from_location(location, None, line_pos, col_pos+1,
                                                            False, False)
                def tab_created():
                    self.window.set_active_tab(tab)
                    view = tab.get_view()
                    GObject.idle_add(view.grab_focus)
                GObject.idle_add(tab_created)
            else:
                view = self.set_active_tab(tab, line_pos, col_pos)
                GObject.idle_add(view.grab_focus)
            
    def set_active_tab(self, tab, lineno, offset):
        self.window.set_active_tab(tab)
        view = tab.get_view()
        if lineno > 0:
            doc = tab.get_document()
            doc.goto_line(lineno - 1)
            cur_iter = doc.get_iter_at_line(lineno-1)
            linelen = cur_iter.get_chars_in_line() - 1
            if offset >= linelen:
                cur_iter.forward_to_line_end()
            elif offset >= 0:
                cur_iter.set_line_offset(offset)
            doc.place_cursor(cur_iter)
            view.scroll_to_cursor()
        return view
        
        
