#!/usr/bin/python
# -*- coding: utf-8 -*-

#  Copyright © 2009-2012  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

_globals = globals().copy()

import sys
import os


# internal functions for color output
YELLOW  = lambda s: '\033[1;33m%s\033[0;37m' % s
RED     = lambda s: '\033[1;31m%s\033[0;37m' % s
WHITE   = lambda s: '\033[1;37m%s\033[0m' % s
BLUE    = lambda s: '\033[1;34m%s\033[0m' % s
LBLUE   = lambda s: '\033[1;36m%s\033[0m' % s
ULEN = len(YELLOW('')) * 2 + len(WHITE('')) + 77

_showwarning = None

def _print_color_line(filename_cwd, filename, lineno, func_name, text):
    filename_ = filename
    if filename_cwd and filename.startswith(filename_cwd):
        filename_ = filename[len(filename_cwd):].lstrip('/')
    line1 = '%s: %s in %s' % (YELLOW(filename_), RED(lineno), WHITE(func_name))
    if text:
        if len(line1) + len(text) <= ULEN:
            print('%s: %s' % (line1, text), file=sys.__stderr__)
        else:
            print(line1, file=sys.__stderr__)
            print(' ', text[:77], file=sys.__stderr__)
    else:
        print(line1, file=sys.__stderr__)

def _print_color_lines(exc_type, exc_value, tb):
    try:
        filename_cwd = os.getcwd()
    except OSError:
        filename_cwd = None
    print('-'*20, 'Traceback', '-'*20, file=sys.__stderr__)
    for filename, lineno, func_name, text in tb:
        if filename == '<string>' and lineno <= 1 and func_name == '<lambda>' and not text:
            continue
        _print_color_line(filename_cwd, filename, lineno, func_name, text)
    try:
        exc_value = LBLUE(exc_value)
    except Exception:
        exc_value = LBLUE('<unprintable object>')
    print('%s: %s' % (BLUE(exc_type.__name__), exc_value), file=sys.__stderr__)
    

# internal functions for handling exceptions
def _send_dbus_tb(exc_type, exc_value, tb):
    filename = ''
    lineno = 0
    offset = 0
    tb_out_list = []
    
    for filename, lineno, func_name, text in tb:
        if os.path.exists(filename):
            filename = os.path.realpath(filename)
        tb_out_list.append( (filename, lineno or 0, func_name, text or '') )
    
    if issubclass(exc_type, SyntaxError):
        #print 'SyntaxError found' 
        filename = exc_value.filename
        if filename and os.path.exists(filename):
            filename = os.path.realpath(filename)
        lineno = exc_value.lineno or 0
        offset = exc_value.offset
    elif exc_type.__name__ == 'DBusException':
        # DBusException can have traceback in args, strip it,
        #   exception should be send on the other side
        exc_value_args = list(exc_value.args)
        exc_value_args[0] = exc_value_args[0].rstrip('\n').split('\n')[-1]
        exc_value.args = tuple(exc_value_args)
        
    try:
        exc_value = str(exc_value)
    except Exception:
        exc_value = '<unprintable object>'
    send_dbus_message(exc_type.__name__, exc_value, filename, lineno, offset, tb_out_list)
    

# exception hooks
def _color_excepthook(exc_type, exc_value, exc_traceback):
    import traceback
    tb = traceback.extract_tb(exc_traceback)
    _print_color_lines(exc_type, exc_value, tb)
    
def _dbus_excepthook(exc_type, exc_value, exc_traceback):
    import traceback
    tb = traceback.extract_tb(exc_traceback)
    sys.__excepthook__(exc_type, exc_value, exc_traceback)
    _send_dbus_tb(exc_type, exc_value, tb)
    
def _all_excepthook(exc_type, exc_value, exc_traceback):
    import traceback
    tb = traceback.extract_tb(exc_traceback)
    _print_color_lines(exc_type, exc_value, tb)
    _send_dbus_tb(exc_type, exc_value, tb)
    

# internal functions for handling warnings
def _print_color_warning(category, message, filename, lineno, text):
    filename_cwd = os.getcwd()
    if filename.startswith(filename_cwd):
        filename = filename[len(filename_cwd):].lstrip('/')
    print('-'*20, 'Warning', '-'*20, file=sys.__stderr__)
    line1 = '%s: %s' % (YELLOW(filename), RED(lineno))
    if text:
        if len(line1) + len(text) <= ULEN:
            print('%s: %s' % (line1, text), file=sys.__stderr__)
        else:
            print(line1, file=sys.__stderr__)
            print(' ', text[:77], file=sys.__stderr__)
    else:
        print(line1, file=sys.__stderr__)
    print('%s: %s' % (BLUE(category.__name__), LBLUE(message)), file=sys.__stderr__)

def _send_dbus_warning(message, category, filename, lineno):
    filename = os.path.abspath(filename)
    send_dbus_message(category.__name__, str(message), filename, lineno)
    

# warning hooks
def _color_warninghook(message, category, filename, lineno, unused_file_=None, line=None):
    _print_color_warning(category, message, filename, lineno, line)

def _dbus_warninghook(message, category, filename, lineno, file_=None, line=None):
    import warnings
    _showwarning(message, category, filename, lineno, file_, line)
    _send_dbus_warning(message, category, filename, lineno)

def _all_warninghook(message, category, filename, lineno, unused_file_=None, line=None):
    _print_color_warning(category, message, filename, lineno, line)
    _send_dbus_warning(message, category, filename, lineno)

def _send_dbus_command(command, timestamp):
    import subprocess
    if timestamp is None:
        import time
        timestamp = time.time()
    command = str((timestamp, command))
    if type(command) is not bytes:
        command = bytes(command, 'utf-8')
    send_cmd_path = os.path.join(os.path.dirname(__file__), 'send-traceback.py')
    p = subprocess.Popen(('python', send_cmd_path),
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT)
    p.stdin.write(command)
    p.stdin.close()
    

# public API
def send_dbus_message(category, description, filename='', lineno=1, offset=0,
                        submessages=[], timestamp=None):
    '''Send messages over DBus
    
    category: one word that describes the type of the message,
              i.e. error, warning, NameError, funcname
    description: text that explains the message
    filename, lineno, offset: position in a file
    submessages: list of tuples, each tuple contains four elements:
        filename, lineno, category, description
    timestamp: float, messages are ordered according to this value
               None, time.time() is used
    '''
    message = str((category, description, submessages, filename, lineno, offset))
    _send_dbus_command(message, timestamp)
    
def send_dbus_clear(timestamp=None):
    '''Send CLEAR command'''
    _send_dbus_command('CLEAR', timestamp)
    
def _set_hooks(excepthook, warninghook, showwarning):
    if showwarning:
        global _showwarning
        import warnings
        if _showwarning is None:
            _showwarning = warnings.showwarning
        warnings.showwarning = warninghook
    sys.excepthook = excepthook
    
def install_all(showwarning=True):
    try:
        import fcntl, termios
        fcntl.ioctl(1, termios.TIOCGWINSZ, '1234')
    except:
        _set_hooks(_dbus_excepthook, _dbus_warninghook, showwarning)
    else:
        _set_hooks(_all_excepthook, _all_warninghook, showwarning)
    
def install_dbus(showwarning=True):
    _set_hooks(_dbus_excepthook, _dbus_warninghook, showwarning)

def install_color(showwarning=True):
    try:
        import fcntl, termios
        fcntl.ioctl(1, termios.TIOCGWINSZ, '1234')
    except:
        pass
    else:
        _set_hooks(_color_excepthook, _color_warninghook, showwarning)

def install_orig(showwarning=True):
    _set_hooks(sys.__excepthook__, _showwarning, showwarning)
    
install = install_dbus

__all__ = [ 'install_all', 'install_color', 'install_dbus', 'install_orig', 'install',
            'send_dbus_clear', 'send_dbus_message']


#if __name__ == '__main__':
#    install_dbus()
#    del sys.argv[0]
#    sys.path[0] = os.path.dirname(os.path.realpath(sys.argv[0]))
#    _globals['__file__'] = sys.argv[0]
#    exec(open(sys.argv[0]).read(), _globals)

